
/*
 * Program name: ll_get_machine_list.c
 * ---------------------------------------------------------------------------------
 * This program shows how LoadLeveler's Data Access API can be used
 * to obtain information about machines that have been assigned by LoadLeveler
 * to run various "tasks" of a user job step. It is intended for use mostly
 * with MPICH and MPICH-GM job steps. The name of the job step is from
 * the environment variable LOADL_STEP_ID that is set by LoadLeveler before
 * the job is started.  For each task of the job step, the name of the machine
 * assigned to run the task is written to standard output.
 *
 * Input:
 *   - LOADL_STEP_ID environment variable
 *
 * Output:
 *   - stdout: List of machines that have been selected by LoadLeveler
 *             to run the tasks of the job step. One machine per line,
 *             sorted by task instance ID.
 *   - stderr: Error messages.
 *
 * Return code:
 *    = 0: Success. Machine list created.
 *    = 1: No machine has been assigned to the job step.
 *    = 2: All other errors.
 *
 * Compile and link this program with:
 *    g++ -c ll_get_machine_list.c -I/opt/ibmll/LoadL/full/include
 *    g++ -o ll_get_machine_list ll_get_machine_list.o -lllapi
 * g++ must be used instead of gcc since the LoadLeveler libllapi.so library is
 * a C++ library.
 * ---------------------------------------------------------------------------------
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "llapi.h"

void ll_get_data_error(char *, int);
int ll_task_info_compare(const void *, const void *);

typedef struct {
   int   task_instance_id;
   char *task_instance_machine_name;
} LL_TASK_INFO;

#define LL_TASK_INFO_ARRAY_SIZE 5000
LL_TASK_INFO  *ll_task_info_array[LL_TASK_INFO_ARRAY_SIZE];

int
main(int argc, char *argv[])
{
   LL_element *queryObject, *job, *step, *node, *task, *task_instance;
   int i, rc, obj_count, err_code, ll_master_task, job_step_count;
   char *ll_step_id, *job_step_list[2], *task_machine_name;
   int i_value32, count, ll_task_info_count;
   char *schedd_host_name, *ptr;
   LL_TASK_INFO  *ll_task_info;

   /* Get the step ID from LOADL_STEP_ID environment variable. */
   ll_step_id = getenv("LOADL_STEP_ID");
   if (!ll_step_id) {
      fprintf(stderr, "Unable to determine the job step ID: LOADL_STEP_ID env. variable is not defined.\n");
      exit(2);
   }

   job_step_list[0] = ll_step_id;
   job_step_list[1] = NULL;

   /* Get name of LoadL_schedd we want to query from the name of the job step. */
   /* Job step ID has format: fully_qualified_host_name.job_id.step_id */
   schedd_host_name = strdup(ll_step_id);
   ptr = schedd_host_name + strlen(schedd_host_name);
   count = 0;
   for (; ptr > schedd_host_name ; ptr--) {
      if (*ptr == '.') count++;
      if (count == 2) {
         *ptr = '\0';
         break;
      }
   }
   if (count != 2) {
      fprintf(stderr, "Job step name associated with the LOADL_STEP_ID env. variable is not a valid step name.\n");
      exit(2);
   }

   /* Initialize the LL API. Specify that query type is JOBS. */
   queryObject = ll_query(JOBS);
   if (!queryObject) {
      fprintf(stderr, "Query JOBS Error: ll_query() returns NULL.\n");
      exit(2);
   }

   /* Specify that this is a QUERY_STEPID type of query. */
   rc = ll_set_request(queryObject, QUERY_STEPID, job_step_list, ALL_DATA);
   if (rc) {
      fprintf(stderr, "ll_set_request() for QUERY_STEPID Error. RC = %d.\n", rc);
      exit(2);
   }

   /* Get a Job object from LoadL_schedd that contains the relevant job step. */
   job = ll_get_objs(queryObject, LL_SCHEDD, schedd_host_name, &obj_count, &err_code);
   if (!job) {
      fprintf(stderr, "ll_get_objs() returns NULL. Error code = %d\n", err_code);
      exit(2);
   }

   if (obj_count != 1) {  /* Only 1 Job object is expected. */
      fprintf(stderr, "ll_get_objs() Error: An unexpected number of job object count is returned.\n");
      exit(2);
   }

   rc = ll_get_data(job, LL_JobStepCount, &job_step_count);
   if (rc) ll_get_data_error("LL_JobStepCount", rc);
   if (job_step_count != 1) { /* Only 1 Job Step object is expected. */
      fprintf(stderr, "ll_get_data() Error: An unexpected number of job step count is returned.\n");
      exit(2);
   }

   step = NULL;
   rc = ll_get_data(job, LL_JobGetFirstStep, &step);
   if (rc) ll_get_data_error("LL_JobGetFirstStep", rc);
   if (!step) {
      fprintf(stderr, "ll_get_data() Error: Unable to obtain Job Step information.\n");
      exit(2);
   }

   node = NULL;
   rc = ll_get_data(step, LL_StepGetFirstNode, &node);
   if (rc) ll_get_data_error("LL_StepGetFirstNode", rc);
   ll_task_info_count = 0;

   while(node) {     /* Loop through the "Node" objects. */
      task = NULL;
      rc = ll_get_data(node, LL_NodeGetFirstTask, &task);
      if (rc) ll_get_data_error("LL_NodeGetFirstTask", rc);

      while(task) {  /* Loop through the "Task" objects. */
         ll_master_task = 0;
         rc = ll_get_data(task, LL_TaskIsMaster, &ll_master_task);
         if (rc) ll_get_data_error("LL_TaskIsMaster", rc);

         /* The "master task" Task object is a LoadLeveler abstraction and is not relevant here. */
         /* Look at only Task objects that are not "master". */
         if (!ll_master_task) {
            task_instance = NULL;
            rc = ll_get_data(task, LL_TaskGetFirstTaskInstance, &task_instance);
            if (rc) ll_get_data_error("LL_TaskGetFirstTaskInstance", rc);

            while (task_instance) {  /* Loop through the "Task Instance" objects. */
               task_machine_name = NULL;
               rc = ll_get_data(task_instance, LL_TaskInstanceMachineName, &task_machine_name);
               if (rc) ll_get_data_error("LL_TaskInstanceMachineName", rc);
               else {
                  if (task_machine_name) {
                     if (ll_task_info_count > (LL_TASK_INFO_ARRAY_SIZE -1)) {
                        fprintf(stderr, "Increase the value of LL_TASK_INFO_ARRAY_SIZE and recompile program.\n");
                        exit(2);
                     }
                     /* Store machine name and associated task Instance ID in a LL_TASK_INFO structure. */ 
                     ll_task_info = (LL_TASK_INFO *) malloc(sizeof(LL_TASK_INFO));
                     if (!ll_task_info) {
                        fprintf(stderr, "Unable to allocate memory: malloc() returns NULL.\n");
                        exit(2);
                     }
                     ll_task_info->task_instance_machine_name = strdup(task_machine_name);
                     free(task_machine_name);
                     rc = ll_get_data(task_instance, LL_TaskInstanceTaskID, &i_value32);
                     if (rc) ll_get_data_error("LL_TaskInstanceTaskID", rc);
                     ll_task_info->task_instance_id = i_value32;
                     ll_task_info_array[ll_task_info_count] = ll_task_info;
                     ll_task_info_count++;
                  }
               }
               task_instance = NULL;
               rc = ll_get_data(task, LL_TaskGetNextTaskInstance, &task_instance);
               if (rc) ll_get_data_error("LL_TaskGetNextTaskInstance", rc);
            }
         }
         task = NULL;
         rc = ll_get_data(node, LL_NodeGetNextTask, &task);
         if (rc) ll_get_data_error("LL_NodeGetNextTask", rc);
      }
      node = NULL;
      rc = ll_get_data(step, LL_StepGetNextNode, &node);
      if (rc) ll_get_data_error("LL_StepGetNextNode", rc);
   }

   if (ll_task_info_count > 0) {
      /* Sort the array of pointers to LL_TASK_INFO structures by task instance ID */
      qsort((void *) ll_task_info_array, ll_task_info_count, sizeof(LL_TASK_INFO *), ll_task_info_compare);

      /* Write sorted machine list to standard output. */
      for (i = 0; i < ll_task_info_count; i++) {
         fprintf(stdout, "%s\n", ll_task_info_array[i]->task_instance_machine_name);
      }
   }

   ll_free_objs(queryObject);
   ll_deallocate(queryObject);
   if (ll_task_info_count < 1) return (1);
   return (0);
}

int
ll_task_info_compare(const void *p1, const void *p2)
{
   LL_TASK_INFO **ptr1, **ptr2;
   ptr1 = (LL_TASK_INFO **) p1;
   ptr2 = (LL_TASK_INFO **) p2;
   if ((*ptr1)->task_instance_id < (*ptr2)->task_instance_id) return (-1);
   if ((*ptr1)->task_instance_id > (*ptr2)->task_instance_id) return (+1);
   return (0);
}

void
ll_get_data_error(char *ptr, int rc)
{
   fprintf(stderr, "Error: ll_get_data() for %s failed. Return Code = %d\n", ptr, rc);
   exit(2);
}

